"""SEG-Y parser configuration."""

from __future__ import annotations

from typing import Any

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

from segy.schema import Endianness


class SegyBaseSettings(BaseSettings):
    """Base class for settings."""

    model_config = SettingsConfigDict(extra="ignore", case_sensitive=False)


class SegyFieldSetting(SegyBaseSettings):
    """Setting class to configure a field (key or override)."""

    key: str = Field(...)
    value: int | None = Field(...)


class SamplesPerTraceSetting(SegyFieldSetting):
    """Configuration for samples per trace parsing."""

    key: str = "samples_per_trace"
    value: int | None = None


class SampleIntervalSetting(SegyFieldSetting):
    """Configuration for samples interval parsing."""

    key: str = "sample_interval"
    value: int | None = None


class ExtendedTextHeaderSetting(SegyFieldSetting):
    """Configuration for extended textual headers parsing."""

    key: str = "extended_textual_headers"
    value: int | None = None


class SegyBinaryHeaderSettings(SegyBaseSettings):
    """SEG-Y binary header parsing settings."""

    samples_per_trace: SamplesPerTraceSetting = SamplesPerTraceSetting()
    sample_interval: SampleIntervalSetting = SampleIntervalSetting()
    extended_text_header: ExtendedTextHeaderSetting = ExtendedTextHeaderSetting()


class SegyFileSettings(SegyBaseSettings):
    """SEG-Y file parsing settings."""

    binary: SegyBinaryHeaderSettings = Field(default_factory=SegyBinaryHeaderSettings)
    endianness: Endianness = Field(default=Endianness.BIG)
    revision: int | float | None = Field(default=None)

    storage_options: dict[str, Any] = Field(default_factory=dict)
    apply_transforms: bool = Field(default=True)

    model_config = SettingsConfigDict(
        env_prefix="SEGY__",
        env_nested_delimiter="__",
    )



"""Custom array interface.

We subclass NumPy ndarray with some methods to enrich it for
better use experience. Like dictionary or JSON dumps for
structured arrays.

See here for details:
https://numpy.org/doc/stable/user/basics.subclassing.html
"""

from __future__ import annotations

from json import dumps as json_dumps
from typing import TYPE_CHECKING

import numpy as np
from pandas import DataFrame

if TYPE_CHECKING:
    from typing import Any

    from numpy.typing import NDArray


class SegyArray(np.ndarray):  # type: ignore[type-arg]
    """Base class for array interface. Like ndarray but extensible."""

    def __new__(cls, input_array: NDArray[Any]) -> SegyArray:
        """Numpy subclass logic."""
        return np.asarray(input_array).view(cls)

    def __array_finalize__(self, obj: NDArray[Any] | None) -> None:
        """Numpy subclass logic."""
        if obj is None:
            return


class HeaderArray(SegyArray):
    """Header ndarray with convenience features."""

    def to_dict(self) -> dict[str, Any]:
        """Convert header to dict."""
        result_dict = {}

        if self.dtype.names is None:
            msg = f"{self.__class__.__name__} can only work on structured arrays."
            raise ValueError(msg)

        for field in self.dtype.names:
            field_values = self[field]
            result_dict[field] = field_values.squeeze().tolist()

        return result_dict

    def to_json(self, indent: int = 2) -> str:
        """Convert header to JSON."""
        return json_dumps(self.to_dict(), indent=indent)

    def to_dataframe(self) -> DataFrame:
        """Convert structured data to pandas DataFrame."""
        return DataFrame.from_records(self)


class TraceArray(SegyArray):
    """Trace ndarray with convenience features."""

    @property
    def header(self) -> HeaderArray:
        """Access headers of the trace(s)."""
        return HeaderArray(self["header"])

    @property
    def sample(self) -> NDArray[Any]:
        """Access data of the trace(s)."""
        return self["sample"]



"""Factory methods for SEG-Y file creation."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from segy.schema import Endianness
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from typing import Any

    from numpy.typing import NDArray

    from segy.schema import SegyDescriptor


DEFAULT_TEXT_HEADER_LINES = [
    "C01 File written by the open-source segy library.",
    "C02",
    "C03 Website: https://segy.readthedocs.io",
    "C04 Source: https://github.com/TGSAI/segy",
]

DEFAULT_TEXT_HEADER_LINES += [f"C{line_no:02}" for line_no in range(5, 40)]
DEFAULT_TEXT_HEADER_LINES += ["C40 END TEXTUAL HEADER"]
DEFAULT_TEXT_HEADER_LINES = [line.ljust(80) for line in DEFAULT_TEXT_HEADER_LINES]
DEFAULT_TEXT_HEADER = "\n".join(DEFAULT_TEXT_HEADER_LINES)


class SegyFactory:
    """Factory class for composing SEG-Y by components.

    Args:
        spec: SEG-Y specification.
        sample_interval: Sample interval to use in new file.
        samples_per_trace: Number of samples per trace.
    """

    def __init__(
        self,
        spec: SegyDescriptor,
        sample_interval: int = 4000,
        samples_per_trace: int = 1500,
    ) -> None:
        self.spec = spec

        self.sample_interval = sample_interval
        self.samples_per_trace = samples_per_trace

        self.spec.trace.sample_descriptor.samples = samples_per_trace

    def create_textual_header(self, text: str | None = None) -> bytes:
        """Create a textual header for the SEG-Y file.

        The length of the text should match the rows and columns in the spec's
        TextHeaderDescriptor. The newlines must also be in the text to separate
        the rows.

        Args:
            text: String containing text header rows. If left as None, a default
                textual header will be created.

        Returns:
            Bytes containing the encoded text header, ready to write.
        """
        text = DEFAULT_TEXT_HEADER if text is None else text

        text_descriptor = self.spec.text_file_header
        text = text_descriptor._unwrap(text)

        return text_descriptor._encode(text)

    def create_binary_header(self) -> bytes:
        """Create a binary header for the SEG-Y file.

        Returns:
            Bytes containing the encoded binary header, ready to write.
        """
        binary_descriptor = self.spec.binary_file_header
        bin_header = np.zeros(shape=1, dtype=binary_descriptor.dtype)

        bin_header["seg_y_revision"] = self.spec.segy_standard.value * 256
        bin_header["sample_interval"] = self.sample_interval
        bin_header["sample_interval_orig"] = self.sample_interval
        bin_header["samples_per_trace"] = self.samples_per_trace
        bin_header["samples_per_trace_orig"] = self.samples_per_trace

        return bin_header.tobytes()

    def create_trace_header_template(
        self,
        size: int = 1,
        fill: bool = True,
    ) -> NDArray[Any]:
        """Create a trace header template array that conforms to the SEG-Y spec.

        Args:
            size: Number of headers for the template.
            fill: Optional, fill with zeros. Default is True.

        Returns:
            Array containing the trace header template.
        """
        descriptor = self.spec.trace.header_descriptor
        dtype = descriptor.dtype.newbyteorder(Endianness.NATIVE.symbol)

        header_template = np.empty(shape=size, dtype=dtype)

        if fill is True:
            header_template.fill(0)

        header_template["sample_interval"] = self.sample_interval
        header_template["samples_per_trace"] = self.samples_per_trace

        return header_template

    def create_trace_data_template(
        self,
        size: int = 1,
        fill: bool = True,
    ) -> NDArray[Any]:
        """Create a trace data template array that conforms to the SEG-Y spec.

        Args:
            size: Number of traces for the template.
            fill: Optional, fill with zeros. Default is True.

        Returns:
            Array containing the trace data template.
        """
        descriptor = self.spec.trace.sample_descriptor
        dtype = descriptor.dtype.newbyteorder(Endianness.NATIVE.symbol)

        data_template = np.empty(shape=size, dtype=dtype)

        if fill is True:
            data_template.fill(0)

        return data_template

    def create_traces(self, headers: NDArray[Any], samples: NDArray[Any]) -> bytes:
        """Convert trace data and header to bytes conforming to SEG-Y spec.

        The rows (length) of the headers and traces must match. The headers
        must be a (num_traces,) shape array and data must be a
        (num_traces, num_samples) shape array. They can be created via the
        `create_trace_header_template` and `create_trace_data_template` methods.

        Args:
            headers: Header array.
            samples: Data array.

        Returns:
            Bytes containing the encoded traces, ready to write.

        Raises:
            AttributeError: if data dimensions are wrong (not 2D trace,samples).
            ValueError: if there is a shape mismatch between headers.
            ValueError: if there is a shape mismatch number of samples.
        """
        trace_descriptor = self.spec.trace
        trace_descriptor.sample_descriptor.samples = self.samples_per_trace

        if samples.ndim != 2:  # noqa: PLR2004
            msg = (
                "Data array must be 2-dimensional with rows as traces "
                "and columns as data samples."
            )
            raise AttributeError(msg)

        if samples.shape[1] != self.samples_per_trace:
            msg = f"Trace length must be {self.samples_per_trace}."
            raise ValueError(msg)

        if len(headers) != len(samples):
            msg = "Header array must have the same number of rows as data array."
            raise ValueError(msg)

        target_endian = trace_descriptor.endianness

        header_pipeline = TransformPipeline()
        data_pipeline = TransformPipeline()

        byte_swap = TransformFactory.create("byte_swap", target_endian)
        ibm_float = TransformFactory.create("ibm_float", "to_ibm")

        header_pipeline.add_transform(byte_swap)
        data_pipeline.add_transform(ibm_float)
        data_pipeline.add_transform(byte_swap)

        trace = np.empty(shape=len(samples), dtype=trace_descriptor.dtype)
        trace["header"] = header_pipeline.apply(headers)
        trace["sample"] = data_pipeline.apply(samples)

        return trace.tobytes()



"""Transforms to apply to arrays and structured arrays."""

from __future__ import annotations

import sys
from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from segy.ibm import ibm2ieee
from segy.ibm import ieee2ibm
from segy.schema import Endianness

if TYPE_CHECKING:
    from typing import Any

    from numpy._typing._dtype_like import _DTypeDict
    from numpy.typing import DTypeLike
    from numpy.typing import NDArray


def get_endianness(data: NDArray[Any]) -> Endianness:
    """Map the numpy byte order character to Endianness."""
    if data.dtype.isnative:
        return Endianness[sys.byteorder.upper()]

    return Endianness.LITTLE if data.dtype.byteorder == "<" else Endianness.BIG


def _extract_nested_dtype(dtype: np.dtype[Any]) -> DTypeLike:
    """Extract nested dtype information from struct dtype."""
    if dtype.names is None or dtype.fields is None:
        return dtype

    dtype_info: _DTypeDict = {
        "names": dtype.names,
        "formats": [
            _extract_nested_dtype(dtype.fields[name][0]) for name in dtype.names
        ],
        "offsets": [dtype.fields[name][1] for name in dtype.names],
        "itemsize": dtype.itemsize,
    }

    return dtype_info


def _modify_dtype_field(
    old_dtype: np.dtype[np.void],
    key: str,
    new_type: np.dtype[Any],
) -> np.dtype[np.void]:
    """Constructs a new dtype for the structured array after a type change in one of its fields."""
    if old_dtype.names is None:  # pragma: no cover
        msg = "Cannot modify dtype for non-structured array."
        raise ValueError(msg)

    dtype_info: _DTypeDict = _extract_nested_dtype(old_dtype)  # type: ignore[assignment]

    key_index = dtype_info["names"].index(key)
    key_to_modify = dtype_info["formats"][key_index]

    if key_to_modify.kind == "V":
        new_type = np.dtype((new_type.str,) + key_to_modify.subdtype[1:])

    dtype_info["formats"][key_index] = new_type  # type: ignore[index]

    # Check if field size changed
    old_size = key_to_modify.itemsize
    new_size = new_type.itemsize
    diff_size = new_size - old_size

    if diff_size != 0:
        offsets = dtype_info["offsets"]
        itemsize = dtype_info["itemsize"]
        dtype_info["offsets"] = [
            x + (diff_size if i > key_index else 0) for i, x in enumerate(offsets)
        ]
        dtype_info["itemsize"] = itemsize + diff_size

    return np.dtype(dtype_info)


def _modify_structured_field(
    data: NDArray[np.void],
    key: str,
    transformed: NDArray[Any],
) -> NDArray[np.void]:
    """Handle structured array assignment with dtype change or promotion."""
    if data.dtype.names is None:  # pragma: no cover
        msg = "Cannot modify dtype for non-structured array."
        raise ValueError(msg)

    # Assign and return early if basic types are the same. Numpy handles endianness.
    if transformed.dtype == data[key].dtype:
        data[key] = transformed
        return data

    # If transform changed field type, we need to define the new array dtype.
    old_field_dtype = data[key].dtype
    new_field_dtype = transformed.dtype
    new_struct_dtype = _modify_dtype_field(data.dtype, key, new_field_dtype)

    # If the field size hasn't changed we can do it in-place with view magic, no copy.
    # Example: uint32 scaled by a negative integer, result fits into output: int32.
    if new_struct_dtype.itemsize == data.dtype.itemsize:
        data[key] = transformed.view(old_field_dtype)
        return data.view(new_struct_dtype)

    # Worst case scenario, field size changes. We need to initialize a new
    # array and copy all the old data and the new data to appropriate fields.
    # Example: int32 field scaled by float32 scalar, resulting in float64 output.
    new_data = np.empty_like(data, dtype=new_struct_dtype)
    for orig_key in data.dtype.names:
        if orig_key == key:  # skip transformed key
            continue
        new_data[orig_key] = data[orig_key]

    new_data[key] = transformed

    return new_data


class Transform:
    """Base class for header transformation strategies."""

    def __init__(self, keys: list[str] | None = None) -> None:
        self.keys = keys

    def apply(self, data: NDArray[Any]) -> NDArray[Any]:
        """Applies transformation based on ndarray or struct array."""
        if self.keys is None:
            return self._transform(data)

        return self._transform_struct(data)

    @abstractmethod
    def _transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Abstract transformation method on numpy arrays."""

    def _transform_struct(self, data: NDArray[Any]) -> NDArray[Any]:
        """Generalized transform for structured arrays."""
        if self.keys is None:  # pragma: no cover
            msg = "Trying to modify structured array fields with no keys provided."
            raise ValueError(msg)

        if data.dtype.names is None:
            msg = "The array to transform doesn't contain any fields."
            raise ValueError(msg)

        missing_fields = set(self.keys) - set(data.dtype.names)
        if missing_fields:
            msg = f"Field(s) {missing_fields} not in the array."
            raise ValueError(msg)

        for key in self.keys:
            transformed = self._transform(data[key])
            data = _modify_structured_field(data, key, transformed)

        return data


class ScaleTransform(Transform):
    """Scales numeric data by a specified factor.

    Args:
        scalar: Scalar to apply to the data.
        keys: Optional list of keys to apply the transform.
    """

    def __init__(self, scalar: float, keys: list[str] | None = None) -> None:
        super().__init__(keys)
        self.scalar = scalar

    def _transform(self, data: NDArray[Any]) -> NDArray[Any]:
        orig_endian = get_endianness(data)
        data = data * self.scalar
        new_endian = get_endianness(data)
        if new_endian is not orig_endian:
            data = data.byteswap().newbyteorder()
        return data


class ByteSwapTransform(Transform):
    """Byte swaps numeric data by based on target order.

    Args:
        target_order: Desired byte order.
    """

    def __init__(self, target_order: Endianness) -> None:
        super().__init__()
        self.target_order = target_order

    def _transform(self, data: NDArray[Any]) -> NDArray[Any]:
        source_order = get_endianness(data)

        if source_order != self.target_order:
            data = data.newbyteorder(self.target_order.symbol).byteswap()

        return data


class CastTypeTransform(Transform):
    """Casts numeric data based on target type.

    Args:
        target_type: Desired data type to cast.
        keys: Optional list of keys to apply the transform.
    """

    def __init__(self, target_type: DTypeLike, keys: list[str] | None = None) -> None:
        super().__init__(keys=keys)
        self.target_type = np.dtype(target_type)

    def _transform(self, data: NDArray[Any]) -> NDArray[Any]:
        return data.astype(self.target_type)


class IbmFloatTransform(Transform):
    """IBM float convert an array.

    Args:
        direction: IBM Float conversion direction.
        keys: Optional list of keys to apply the transform.
    """

    ibm_func_map = {
        "to_ibm": lambda x: ieee2ibm(x.astype("float32")),
        "to_ieee": lambda x: ibm2ieee(x.view("uint32")),
    }

    def __init__(self, direction: str, keys: list[str] | None = None) -> None:
        super().__init__(keys)
        self.direction = direction

    def _transform(self, data: NDArray[Any]) -> NDArray[Any]:
        return self.ibm_func_map[self.direction](data)  # type: ignore


class TransformFactory:
    """Factory class to generate transformation strategies."""

    transform_map: dict[str, type[Transform]] = {
        "scale": ScaleTransform,
        "byte_swap": ByteSwapTransform,
        "ibm_float": IbmFloatTransform,
        "cast": CastTypeTransform,
    }

    @classmethod
    def create(
        cls: type[TransformFactory],
        transform_type: str,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> Transform:
        """Create an instance of transformation and return it."""
        if transform_type not in cls.transform_map:
            msg = f"Unsupported transformation type: {transform_type}"
            raise KeyError(msg)

        return cls.transform_map[transform_type](*args, **kwargs)


class TransformPipeline:
    """Executes a sequence of transformations on data."""

    def __init__(self) -> None:
        self.transforms: list[Transform] = []

    def add_transform(self, transform: Transform) -> None:
        """Adds a transformation to the pipeline."""
        self.transforms.append(transform)

    def apply(self, data: NDArray[Any]) -> NDArray[Any]:
        """Applies all transformations in sequence."""
        for transform in self.transforms:
            data = transform.apply(data)
        return data



"""Data models for manipulating all kinds of SEG-Y files."""

from segy.schema.data_type import Endianness
from segy.schema.data_type import ScalarType
from segy.schema.data_type import StructuredDataTypeDescriptor
from segy.schema.data_type import StructuredFieldDescriptor
from segy.schema.header import TextHeaderDescriptor
from segy.schema.header import TextHeaderEncoding
from segy.schema.segy import SegyDescriptor
from segy.schema.segy import SegyStandard
from segy.schema.trace import TraceDescriptor
from segy.schema.trace import TraceSampleDescriptor

__all__ = [
    "Endianness",
    "ScalarType",
    "StructuredDataTypeDescriptor",
    "StructuredFieldDescriptor",
    "TextHeaderDescriptor",
    "TextHeaderEncoding",
    "TraceSampleDescriptor",
    "TraceDescriptor",
    "SegyDescriptor",
    "SegyStandard",
]



"""SEG-Y file in-memory constructors."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from fsspec.core import url_to_fs

from segy.arrays import HeaderArray
from segy.config import SegyFileSettings
from segy.indexing import HeaderIndexer
from segy.indexing import SampleIndexer
from segy.indexing import TraceIndexer
from segy.schema import Endianness
from segy.schema import ScalarType
from segy.schema import SegyStandard
from segy.standards.registry import get_spec
from segy.standards.rev1 import rev1_binary_file_header
from segy.standards.rev1 import rev1_extended_text_header
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem
    from numpy.typing import NDArray

    from segy.indexing import AbstractIndexer
    from segy.schema import SegyDescriptor


class SegyFile:
    """A SEG-Y file class that has various accessors.

    Args:
        url: Path to SEG-Y file on disk or remote store.
        spec: The schema / spec describing the SEG-Y file. This
            is optional and by default it will try to infer the
            SEG-Y standard from the binary header.
        settings: A settings instance to configure / override
            the SEG-Y parsing logic. Optional.
    """

    fs: AbstractFileSystem
    url: str

    def __init__(
        self,
        url: str,
        spec: SegyDescriptor | None = None,
        settings: SegyFileSettings | None = None,
    ):
        self.settings = SegyFileSettings() if settings is None else settings

        self.fs, self.url = url_to_fs(url, **self.settings.storage_options)

        self.spec = self._infer_standard() if spec is None else spec

        self._info = self.fs.info(self.url)
        self._update_spec_endianness()
        self._parse_binary_header()

    def _update_spec_endianness(self) -> None:
        """If spec has no endianness, get it from settings."""
        if self.spec.endianness is None:
            self.spec.endianness = self.settings.endianness

    @property
    def file_size(self) -> int:
        """Return file size in bytes."""
        size: int = self._info["size"]
        return size

    @property
    def num_ext_text(self) -> int:
        """Return number of extended text headers."""
        if self.spec.segy_standard in {SegyStandard.REV0, SegyStandard.CUSTOM}:
            return 0

        # Overriding from settings
        if self.settings.binary.extended_text_header.value is not None:
            return self.settings.binary.extended_text_header.value

        header_key = self.settings.binary.extended_text_header.key
        return int(self.binary_header[header_key][0])

    @property
    def samples_per_trace(self) -> int:
        """Return samples per trace in file based on spec."""
        # Overriding from settings
        if self.settings.binary.samples_per_trace.value is not None:
            return self.settings.binary.samples_per_trace.value

        header_key = self.settings.binary.samples_per_trace.key
        return int(self.binary_header[header_key].item())

    @property
    def sample_interval(self) -> int:
        """Return samples interval in file based on spec."""
        # Overriding from settings
        if self.settings.binary.sample_interval.value is not None:
            return self.settings.binary.sample_interval.value

        header_key = self.settings.binary.sample_interval.key
        return int(self.binary_header[header_key].item())

    @property
    def sample_labels(self) -> NDArray[np.int32]:
        """Return sample axis labels."""
        max_samp = self.sample_interval * self.samples_per_trace
        return np.arange(0, max_samp, self.sample_interval, dtype="int32")

    @property
    def num_traces(self) -> int:
        """Return number of traces in file based on size and spec."""
        file_textual_hdr_size = self.spec.text_file_header.itemsize
        file_bin_hdr_size = self.spec.binary_file_header.itemsize
        trace_size = self.spec.trace.itemsize

        file_metadata_size = file_textual_hdr_size + file_bin_hdr_size

        if self.num_ext_text > 0:
            self.spec.extended_text_header = rev1_extended_text_header

            ext_text_size = self.spec.extended_text_header.itemsize * self.num_ext_text
            file_metadata_size = file_metadata_size + ext_text_size

        return (self.file_size - file_metadata_size) // trace_size

    @cached_property
    def text_header(self) -> str:
        """Return textual file header."""
        text_hdr_desc = self.spec.text_file_header

        buffer = self.fs.read_block(
            fn=self.url,
            offset=text_hdr_desc.offset,
            length=text_hdr_desc.itemsize,
        )

        text_header = text_hdr_desc._decode(buffer)
        return text_hdr_desc._wrap(text_header)

    def _infer_standard(self) -> SegyDescriptor:
        if self.settings.revision is None:
            buffer = self.fs.read_block(
                fn=self.url,
                offset=rev1_binary_file_header.offset,
                length=rev1_binary_file_header.item_size,
            )
            revision = np.frombuffer(buffer, offset=300, dtype=">i2", count=1).item()

            # Per SEG-Y standard, there is a Q-point between the bytes. Dividing
            # by 2^8 to get the floating-point value of the revision.
            revision = revision / 256.0

        else:
            revision = self.settings.revision

        standard = SegyStandard(revision)
        return get_spec(standard)

    @cached_property
    def binary_header(self) -> HeaderArray:
        """Read binary header from store, based on spec."""
        buffer = self.fs.read_block(
            fn=self.url,
            offset=self.spec.binary_file_header.offset,
            length=self.spec.binary_file_header.itemsize,
        )

        bin_hdr = np.frombuffer(buffer, dtype=self.spec.binary_file_header.dtype)

        transforms = TransformPipeline()
        transforms.add_transform(
            TransformFactory.create("byte_swap", Endianness.LITTLE),
        )

        return HeaderArray(transforms.apply(bin_hdr))

    def _parse_binary_header(self) -> None:
        """Parse the binary header and apply some rules."""
        # Calculate sizes for dynamic file metadata
        text_hdr_size = self.spec.text_file_header.itemsize
        bin_hdr_size = self.spec.binary_file_header.itemsize

        # Update trace start offset and sample length
        self.spec.trace.sample_descriptor.samples = self.samples_per_trace
        self.spec.trace.offset = text_hdr_size + bin_hdr_size

        if self.num_ext_text > 0:
            self.spec.extended_text_header = rev1_extended_text_header

            ext_text_size = self.spec.extended_text_header.itemsize * self.num_ext_text
            self.spec.trace.offset = self.spec.trace.offset + ext_text_size

    @property
    def sample(self) -> AbstractIndexer:
        """Way to access the file to fetch trace data only."""
        transforms = [
            TransformFactory.create("byte_swap", Endianness.LITTLE),
        ]

        if self.spec.trace.sample_descriptor.format == ScalarType.IBM32:
            transforms.append(TransformFactory.create("ibm_float", "to_ieee"))

        return SampleIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="sample",
            settings=self.settings,
            transforms=transforms,
        )

    @property
    def header(self) -> HeaderIndexer:
        """Way to access the file to fetch trace headers only."""
        transforms = [
            TransformFactory.create("byte_swap", Endianness.LITTLE),
        ]

        return HeaderIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="header",
            settings=self.settings,
            transforms=transforms,
        )

    @property
    def trace(self) -> TraceIndexer:
        """Way to access the file to fetch full traces (headers + data)."""
        transforms = [
            TransformFactory.create("byte_swap", Endianness.LITTLE),
        ]

        if self.spec.trace.sample_descriptor.format == ScalarType.IBM32:
            transforms.append(
                TransformFactory.create("ibm_float", "to_ieee", keys=["sample"])
            )

        return TraceIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="trace",
            settings=self.settings,
            transforms=transforms,
        )



"""Descriptor data model implementations for traces."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from pydantic import Field
from pydantic import model_validator

from segy.schema.base import BaseTypeDescriptor

if TYPE_CHECKING:
    from segy.schema.data_type import Endianness
    from segy.schema.data_type import ScalarType
    from segy.schema.data_type import StructuredDataTypeDescriptor


class TraceSampleDescriptor(BaseTypeDescriptor):
    """A descriptor class for a Trace Samples."""

    format: ScalarType = Field(..., description="Format of trace samples.")  # noqa: A003
    samples: int | None = Field(
        default=None,
        description=(
            "Number of samples in trace. It can be variable, "
            "then it must be read from each trace header."
        ),
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        format_char = self.format.char
        dtype_str = f"{self.samples}{format_char}"
        return np.dtype(dtype_str)


class TraceDescriptor(BaseTypeDescriptor):
    """A descriptor class for a Trace (Header + Data)."""

    header_descriptor: StructuredDataTypeDescriptor = Field(
        ..., description="Trace header descriptor."
    )
    extended_header_descriptor: StructuredDataTypeDescriptor | None = Field(
        default=None, description="Extended trace header descriptor."
    )
    sample_descriptor: TraceSampleDescriptor = Field(
        ..., description="Trace data descriptor."
    )
    offset: int | None = Field(
        default=None, description="Starting offset of the trace."
    )
    endianness: Endianness | None = Field(
        default=None, description="Endianness of traces and headers."
    )

    @model_validator(mode="after")
    def update_submodel_endianness(self) -> TraceDescriptor:
        """Ensure that submodel endianness matches the trace endianness."""
        self.header_descriptor.endianness = self.endianness

        if self.extended_header_descriptor is not None:
            self.extended_header_descriptor.endianness = self.endianness

        return self

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        header_dtype = self.header_descriptor.dtype
        data_dtype = self.sample_descriptor.dtype

        trace_dtype = np.dtype([("header", header_dtype), ("sample", data_dtype)])

        if self.endianness is not None:
            trace_dtype = trace_dtype.newbyteorder(self.endianness.symbol)

        return trace_dtype



"""SEG-Y Revision 0 Specification."""

from segy.schema import ScalarType
from segy.schema import SegyDescriptor
from segy.schema import SegyStandard
from segy.schema import TextHeaderDescriptor
from segy.schema import TextHeaderEncoding
from segy.schema import TraceDescriptor
from segy.schema import TraceSampleDescriptor
from segy.schema.data_type import Endianness
from segy.schema.data_type import StructuredDataTypeDescriptor
from segy.schema.data_type import StructuredFieldDescriptor

BINARY_FILE_HEADER_FIELDS_REV0 = [
    StructuredFieldDescriptor(
        name="job_id",
        offset=0,
        format=ScalarType.INT32,
        description="Job Identification Number",
    ),
    StructuredFieldDescriptor(
        name="line_no",
        offset=4,
        format=ScalarType.INT32,
        description="Line Number",
    ),
    StructuredFieldDescriptor(
        name="reel_no",
        offset=8,
        format=ScalarType.INT32,
        description="Reel Number",
    ),
    StructuredFieldDescriptor(
        name="data_traces_ensemble",
        offset=12,
        format=ScalarType.INT16,
        description="Number of Data Traces per Ensemble",
    ),
    StructuredFieldDescriptor(
        name="aux_traces_ensemble",
        offset=14,
        format=ScalarType.INT16,
        description="Number of Auxiliary Traces per Ensemble",
    ),
    StructuredFieldDescriptor(
        name="sample_interval",
        offset=16,
        format=ScalarType.INT16,
        description="Sample Interval",
    ),
    StructuredFieldDescriptor(
        name="sample_interval_orig",
        offset=18,
        format=ScalarType.INT16,
        description="Sample Interval of Original Field Recording",
    ),
    StructuredFieldDescriptor(
        name="samples_per_trace",
        offset=20,
        format=ScalarType.INT16,
        description="Number of Samples per Data Trace",
    ),
    StructuredFieldDescriptor(
        name="samples_per_trace_orig",
        offset=22,
        format=ScalarType.INT16,
        description="Number of Samples per Data Trace for Original Field Recording",
    ),
    StructuredFieldDescriptor(
        name="data_sample_format",
        offset=24,
        format=ScalarType.INT16,
        description="Data Sample Format Code",
    ),
    StructuredFieldDescriptor(
        name="ensemble_fold",
        offset=26,
        format=ScalarType.INT16,
        description="Ensemble Fold",
    ),
    StructuredFieldDescriptor(
        name="trace_sorting",
        offset=28,
        format=ScalarType.INT16,
        description="Trace Sorting Code",
    ),
    StructuredFieldDescriptor(
        name="vertical_sum",
        offset=30,
        format=ScalarType.INT16,
        description="Vertical Sum Code",
    ),
    StructuredFieldDescriptor(
        name="sweep_freq_start",
        offset=32,
        format=ScalarType.INT16,
        description="Sweep Frequency at Start",
    ),
    StructuredFieldDescriptor(
        name="sweep_freq_end",
        offset=34,
        format=ScalarType.INT16,
        description="Sweep Frequency at End",
    ),
    StructuredFieldDescriptor(
        name="sweep_length",
        offset=36,
        format=ScalarType.INT16,
        description="Sweep Length",
    ),
    StructuredFieldDescriptor(
        name="sweep_type",
        offset=38,
        format=ScalarType.INT16,
        description="Sweep Type Code",
    ),
    StructuredFieldDescriptor(
        name="sweep_trace_no",
        offset=40,
        format=ScalarType.INT16,
        description="Trace Number of Sweep Channel",
    ),
    StructuredFieldDescriptor(
        name="sweep_taper_start",
        offset=42,
        format=ScalarType.INT16,
        description="Sweep Trace Taper Length at Start",
    ),
    StructuredFieldDescriptor(
        name="sweep_taper_end",
        offset=44,
        format=ScalarType.INT16,
        description="Sweep Trace Taper Length at End",
    ),
    StructuredFieldDescriptor(
        name="taper_type",
        offset=46,
        format=ScalarType.INT16,
        description="Taper Type",
    ),
    StructuredFieldDescriptor(
        name="correlated_traces",
        offset=48,
        format=ScalarType.INT16,
        description="Correlated Data Traces",
    ),
    StructuredFieldDescriptor(
        name="binary_gain",
        offset=50,
        format=ScalarType.INT16,
        description="Binary Gain Recovered",
    ),
    StructuredFieldDescriptor(
        name="amp_recovery_method",
        offset=52,
        format=ScalarType.INT16,
        description="Amplitude Recovery Method",
    ),
    StructuredFieldDescriptor(
        name="measurement_system",
        offset=54,
        format=ScalarType.INT16,
        description="Measurement System",
    ),
    StructuredFieldDescriptor(
        name="impulse_signal_polarity",
        offset=56,
        format=ScalarType.INT16,
        description="Impulse Signal Polarity",
    ),
    StructuredFieldDescriptor(
        name="vibratory_polarity",
        offset=58,
        format=ScalarType.INT16,
        description="Vibratory Polarity Code",
    ),
]


TRACE_HEADER_FIELDS_REV0 = [
    StructuredFieldDescriptor(
        name="trace_seq_line",
        offset=0,
        format=ScalarType.INT32,
        description="Trace Sequence Number within Line",
    ),
    StructuredFieldDescriptor(
        name="trace_seq_file",
        offset=4,
        format=ScalarType.INT32,
        description="Trace Sequence Number within File",
    ),
    StructuredFieldDescriptor(
        name="field_rec_no",
        offset=8,
        format=ScalarType.INT32,
        description="Original Field Record Number",
    ),
    StructuredFieldDescriptor(
        name="trace_no_field_rec",
        offset=12,
        format=ScalarType.INT32,
        description="Trace Number within the Field Record",
    ),
    StructuredFieldDescriptor(
        name="energy_src_pt",
        offset=16,
        format=ScalarType.INT32,
        description="Energy Source Point Number",
    ),
    StructuredFieldDescriptor(
        name="cdp_ens_no",
        offset=20,
        format=ScalarType.INT32,
        description="Ensemble Number (CDP, CMP, etc.)",
    ),
    StructuredFieldDescriptor(
        name="trace_no_ens",
        offset=24,
        format=ScalarType.INT32,
        description="Trace Number within the Ensemble",
    ),
    StructuredFieldDescriptor(
        name="trace_id",
        offset=28,
        format=ScalarType.INT16,
        description="Trace Identification Code",
    ),
    StructuredFieldDescriptor(
        name="vert_sum",
        offset=30,
        format=ScalarType.INT16,
        description="Number of Vertically Stacked Traces",
    ),
    StructuredFieldDescriptor(
        name="horiz_stack",
        offset=32,
        format=ScalarType.INT16,
        description="Number of Horizontally Stacked Traces",
    ),
    StructuredFieldDescriptor(
        name="data_use",
        offset=34,
        format=ScalarType.INT16,
        description="Data Use",
    ),
    StructuredFieldDescriptor(
        name="dist_src_to_rec",
        offset=36,
        format=ScalarType.INT32,
        description="Distance from Source Point to Receiver Group",
    ),
    StructuredFieldDescriptor(
        name="rec_elev",
        offset=40,
        format=ScalarType.INT32,
        description="Receiver Group Elevation",
    ),
    StructuredFieldDescriptor(
        name="src_elev",
        offset=44,
        format=ScalarType.INT32,
        description="Source Elevation",
    ),
    StructuredFieldDescriptor(
        name="src_depth",
        offset=48,
        format=ScalarType.INT32,
        description="Source Depth",
    ),
    StructuredFieldDescriptor(
        name="datum_elev_rec",
        offset=52,
        format=ScalarType.INT32,
        description="Datum Elevation at Receiver Group",
    ),
    StructuredFieldDescriptor(
        name="datum_elev_src",
        offset=56,
        format=ScalarType.INT32,
        description="Datum Elevation at Source",
    ),
    StructuredFieldDescriptor(
        name="water_depth_src",
        offset=60,
        format=ScalarType.INT32,
        description="Water Depth at Source",
    ),
    StructuredFieldDescriptor(
        name="water_depth_rec",
        offset=64,
        format=ScalarType.INT32,
        description="Water Depth at Receiver Group",
    ),
    StructuredFieldDescriptor(
        name="scalar_apply_elev",
        offset=68,
        format=ScalarType.INT16,
        description="Scalar to be applied to all elevations and depths",
    ),
    StructuredFieldDescriptor(
        name="scalar_apply_coords",
        offset=70,
        format=ScalarType.INT16,
        description="Scalar to be applied to all coordinates",
    ),
    StructuredFieldDescriptor(
        name="src_x",
        offset=72,
        format=ScalarType.INT32,
        description="Source X coordinate",
    ),
    StructuredFieldDescriptor(
        name="src_y",
        offset=76,
        format=ScalarType.INT32,
        description="Source Y coordinate",
    ),
    StructuredFieldDescriptor(
        name="rec_x",
        offset=80,
        format=ScalarType.INT32,
        description="Receiver X coordinate",
    ),
    StructuredFieldDescriptor(
        name="rec_y",
        offset=84,
        format=ScalarType.INT32,
        description="Receiver Y coordinate",
    ),
    StructuredFieldDescriptor(
        name="coord_units",
        offset=88,
        format=ScalarType.INT16,
        description="Coordinate units",
    ),
    StructuredFieldDescriptor(
        name="weathering_vel",
        offset=90,
        format=ScalarType.INT16,
        description="Weathering Velocity",
    ),
    StructuredFieldDescriptor(
        name="subweathering_vel",
        offset=92,
        format=ScalarType.INT16,
        description="Subweathering Velocity",
    ),
    StructuredFieldDescriptor(
        name="uphole_time_src",
        offset=94,
        format=ScalarType.INT16,
        description="Uphole Time at Source",
    ),
    StructuredFieldDescriptor(
        name="uphole_time_rec",
        offset=96,
        format=ScalarType.INT16,
        description="Uphole Time at Receiver",
    ),
    StructuredFieldDescriptor(
        name="src_static_corr",
        offset=98,
        format=ScalarType.INT16,
        description="Source Static Correction",
    ),
    StructuredFieldDescriptor(
        name="rec_static_corr",
        offset=100,
        format=ScalarType.INT16,
        description="Receiver Static Correction",
    ),
    StructuredFieldDescriptor(
        name="total_static",
        offset=102,
        format=ScalarType.INT16,
        description="Total Static Applied",
    ),
    StructuredFieldDescriptor(
        name="lag_time_a",
        offset=104,
        format=ScalarType.INT16,
        description="Lag Time A",
    ),
    StructuredFieldDescriptor(
        name="lag_time_b",
        offset=106,
        format=ScalarType.INT16,
        description="Lag Time B",
    ),
    StructuredFieldDescriptor(
        name="delay_rec_time",
        offset=108,
        format=ScalarType.INT16,
        description="Delay Recording Time",
    ),
    StructuredFieldDescriptor(
        name="mute_start",
        offset=110,
        format=ScalarType.INT16,
        description="Start Time of Mute",
    ),
    StructuredFieldDescriptor(
        name="mute_end",
        offset=112,
        format=ScalarType.INT16,
        description="End Time of Mute",
    ),
    StructuredFieldDescriptor(
        name="samples_per_trace",
        offset=114,
        format=ScalarType.INT16,
        description="Number of Samples in this Trace",
    ),
    StructuredFieldDescriptor(
        name="sample_interval",
        offset=116,
        format=ScalarType.INT16,
        description="Sample Interval for this Trace",
    ),
    StructuredFieldDescriptor(
        name="gain_type",
        offset=118,
        format=ScalarType.INT16,
        description="Gain Type of Field Instruments",
    ),
    StructuredFieldDescriptor(
        name="instrument_gain",
        offset=120,
        format=ScalarType.INT16,
        description="Instrument Gain Constant",
    ),
    StructuredFieldDescriptor(
        name="instrument_early_gain",
        offset=122,
        format=ScalarType.INT16,
        description="Instrument Early Gain",
    ),
    StructuredFieldDescriptor(
        name="correlated",
        offset=124,
        format=ScalarType.INT16,
        description="Correlated",
    ),
    StructuredFieldDescriptor(
        name="sweep_freq_start",
        offset=126,
        format=ScalarType.INT16,
        description="Sweep Frequency at Start",
    ),
    StructuredFieldDescriptor(
        name="sweep_freq_end",
        offset=128,
        format=ScalarType.INT16,
        description="Sweep Frequency at End",
    ),
    StructuredFieldDescriptor(
        name="sweep_length",
        offset=130,
        format=ScalarType.INT16,
        description="Sweep Length",
    ),
    StructuredFieldDescriptor(
        name="sweep_type",
        offset=132,
        format=ScalarType.INT16,
        description="Sweep Type",
    ),
    StructuredFieldDescriptor(
        name="sweep_trace_taper_start",
        offset=134,
        format=ScalarType.INT16,
        description="Sweep Trace Taper Length at Start",
    ),
    StructuredFieldDescriptor(
        name="sweep_trace_taper_end",
        offset=136,
        format=ScalarType.INT16,
        description="Sweep Trace Taper Length at End",
    ),
    StructuredFieldDescriptor(
        name="taper_type",
        offset=138,
        format=ScalarType.INT16,
        description="Taper Type",
    ),
    StructuredFieldDescriptor(
        name="alias_filter_freq",
        offset=140,
        format=ScalarType.INT16,
        description="Alias Filter Frequency",
    ),
    StructuredFieldDescriptor(
        name="alias_filter_slope",
        offset=142,
        format=ScalarType.INT16,
        description="Alias Filter Slope",
    ),
    StructuredFieldDescriptor(
        name="notch_filter_freq",
        offset=144,
        format=ScalarType.INT16,
        description="Notch Filter Frequency",
    ),
    StructuredFieldDescriptor(
        name="notch_filter_slope",
        offset=146,
        format=ScalarType.INT16,
        description="Notch Filter Slope",
    ),
    StructuredFieldDescriptor(
        name="low_cut_freq",
        offset=148,
        format=ScalarType.INT16,
        description="Low Cut Frequency",
    ),
    StructuredFieldDescriptor(
        name="high_cut_freq",
        offset=150,
        format=ScalarType.INT16,
        description="High Cut Frequency",
    ),
    StructuredFieldDescriptor(
        name="low_cut_slope",
        offset=152,
        format=ScalarType.INT16,
        description="Low Cut Slope",
    ),
    StructuredFieldDescriptor(
        name="high_cut_slope",
        offset=154,
        format=ScalarType.INT16,
        description="High Cut Slope",
    ),
    StructuredFieldDescriptor(
        name="year",
        offset=156,
        format=ScalarType.INT16,
        description="Year Data Recorded",
    ),
    StructuredFieldDescriptor(
        name="day",
        offset=158,
        format=ScalarType.INT16,
        description="Day of Year",
    ),
    StructuredFieldDescriptor(
        name="hour",
        offset=160,
        format=ScalarType.INT16,
        description="Hour of Day",
    ),
    StructuredFieldDescriptor(
        name="minute",
        offset=162,
        format=ScalarType.INT16,
        description="Minute of Hour",
    ),
    StructuredFieldDescriptor(
        name="second",
        offset=164,
        format=ScalarType.INT16,
        description="Second of Minute",
    ),
    StructuredFieldDescriptor(
        name="time_basis_code",
        offset=166,
        format=ScalarType.INT16,
        description="Time Basis Code",
    ),
    StructuredFieldDescriptor(
        name="trace_weighting_factor",
        offset=168,
        format=ScalarType.INT16,
        description="Trace Weighting Factor",
    ),
    StructuredFieldDescriptor(
        name="geophone_group_no_roll1",
        offset=170,
        format=ScalarType.INT16,
        description="Geophone Group Number of Roll Switch Position One",
    ),
    StructuredFieldDescriptor(
        name="geophone_group_no_first_trace",
        offset=172,
        format=ScalarType.INT16,
        description="Geophone Group Number of Trace Number One within Original Field Record",  # noqa: E501
    ),
    StructuredFieldDescriptor(
        name="geophone_group_no_last_trace",
        offset=174,
        format=ScalarType.INT16,
        description="Geophone Group Number of Last Trace within Original Field Record",  # noqa: E501
    ),
    StructuredFieldDescriptor(
        name="gap_size",
        offset=176,
        format=ScalarType.INT16,
        description="Gap Size (total number of groups dropped)",
    ),
    StructuredFieldDescriptor(
        name="over_travel",
        offset=178,
        format=ScalarType.INT16,
        description="Over Travel Associated with Taper",
    ),
]


rev0_textual_file_header = TextHeaderDescriptor(
    rows=40,
    cols=80,
    offset=0,
    encoding=TextHeaderEncoding.EBCDIC,
    format=ScalarType.UINT8,  # noqa: A003
)


rev0_binary_file_header = StructuredDataTypeDescriptor(
    fields=BINARY_FILE_HEADER_FIELDS_REV0,
    item_size=400,
    offset=3200,
)


rev0_trace_header = StructuredDataTypeDescriptor(
    fields=TRACE_HEADER_FIELDS_REV0,
    item_size=240,
)


rev0_trace_data = TraceSampleDescriptor(
    format=ScalarType.IBM32,  # noqa: A003
)


rev0_trace = TraceDescriptor(
    header_descriptor=rev0_trace_header,
    sample_descriptor=rev0_trace_data,
)


rev0_segy = SegyDescriptor(
    segy_standard=SegyStandard.REV0,
    text_file_header=rev0_textual_file_header,
    binary_file_header=rev0_binary_file_header,
    trace=rev0_trace,
    endianness=Endianness.BIG,
)



"""Data models representing data types."""

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import cast

import numpy as np
from pydantic import Field

from segy.compat import StrEnum
from segy.schema.base import BaseTypeDescriptor


class Endianness(StrEnum):
    """Enumeration class with three possible endianness values.

    Examples:
        >>> endian = Endianness.BIG
        >>> print(endian.symbol)
        >
    """

    BIG = "big"
    LITTLE = "little"
    NATIVE = "native"

    def __init__(self, _: str) -> None:
        self._symbol_map = {
            "big": ">",
            "little": "<",
            "native": "=",
        }

    @property
    def symbol(self) -> Literal["<", ">", "="]:
        """Get the numpy symbol for the endianness from mapping."""
        return cast(Literal["<", ">", "="], self._symbol_map[self.value])


class ScalarType(StrEnum):
    """A class representing scalar data types."""

    IBM32 = "ibm32"
    INT64 = "int64"
    INT32 = "int32"
    INT16 = "int16"
    INT8 = "int8"
    UINT64 = "uint64"
    UINT32 = "uint32"
    UINT16 = "uint16"
    UINT8 = "uint8"
    FLOAT64 = "float64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"

    @property
    def char(self) -> str:
        """Returns the numpy character code for a given data type string."""
        if self.value == "ibm32":
            return np.sctype2char("uint32")  # noqa: NPY201

        return np.sctype2char(str(self.value))  # noqa: NPY201


class DataTypeDescriptor(BaseTypeDescriptor):
    """A class representing a descriptor for a data type.

    Examples:
        A float32:

        >>> data_type = DataTypeDescriptor(format="float32")
        >>> data_type.dtype
        dtype('float32')

        A 16-bit unsigned integer:

        >>> data_type = DataTypeDescriptor(format="uint16")
        >>> data_type.dtype
        dtype('uint16')
    """

    format: ScalarType = Field(..., description="The data type of the field.")  # noqa: A003

    @property
    def dtype(self) -> np.dtype[Any]:
        """Converts the byte order and data type of the object into a NumPy dtype."""
        return np.dtype(self.format.char)


class StructuredFieldDescriptor(DataTypeDescriptor):
    """A class representing a descriptor for a structured data-type field.

    Examples:
        A named float at offset 8-bytes:

        >>> data_type = StructuredFieldDescriptor(
        >>>     name="my_var",
        >>>     format="float32",
        >>>     offset=8,
        >>> )

        The name and offset fields will only be used if the structured
        field is used within the context of a :class:`StructuredDataTypeDescriptor`.

        >>> data_type.name
        my_var
        >>> data_type.offset
        8

        The `dtype` property is inherited from :class:`DataTypeDescriptor`.

        >>> data_type.dtype
        dtype('float32')
    """

    name: str = Field(..., description="The short name of the field.")
    offset: int = Field(..., ge=0, description="Starting byte offset.")


class StructuredDataTypeDescriptor(BaseTypeDescriptor):
    """A class representing a descriptor for a structured data-type.

    Examples:
        Let's build a structured data type from scratch!

        We will define three fields with different names, data-types, and
        starting offsets.

        >>> field1 = StructuredFieldDescriptor(
        >>>     name="foo",
        >>>     format="int32",
        >>>     offset=0,
        >>> )
        >>> field2 = StructuredFieldDescriptor(
        >>>     name="bar",
        >>>     format="int16",
        >>>     offset=4,
        >>> )
        >>> field3 = StructuredFieldDescriptor(
        >>>     name="fizz",
        >>>     format="int32",
        >>>     offset=16,
        >>> )

        Note that the fields span the following byte ranges:

        * `field1` between bytes `[0, 4)`
        * `field2` between bytes `[4, 6)`
        * `field3` between bytes `[16, 20)`

        The gap between `field2` and `field3` will be padded with `void`. In
        this case we expect to see an item size of 20-bytes (total length of
        the struct).

        >>> struct_dtype = StructuredDataTypeDescriptor(
        >>>     fields=[field1, field2, field3],
        >>> )

        Now let's look at its data type:

        >>> struct_dtype.dtype
        dtype({'names': ['foo', 'bar', 'fizz'], 'formats': ['<i4', '<i2', '<i4'], 'offsets': [0, 4, 16], 'itemsize': 20})

        If we wanted to pad the end of the struct (to fit a specific byte range),
        we would provide the item_size in the descriptor. If we set it to 30,
        this means that we padded the struct by 10 bytes at the end.

        >>> struct_dtype = StructuredDataTypeDescriptor(
        >>>     fields=[field1, field2, field3],
        >>>     item_size=30,
        >>> )

        Now let's look at its data type:

        >>> struct_dtype.dtype
        dtype({'names': ['foo', 'bar', 'fizz'], 'formats': ['<i4', '<i2', '<i4'], 'offsets': [0, 4, 16], 'itemsize': 30})

        To see what's going under the hood, we can look at a lower level numpy
        description of the `dtype`. Here we observe all the gaps (void types).

        >>> struct_dtype.dtype.descr
        [('foo', '<i4'), ('bar', '<i2'), ('', '|V10'), ('fizz', '<i4'), ('', '|V10')]
    """  # noqa: E501

    fields: list[StructuredFieldDescriptor] = Field(
        ..., description="A list of descriptors for a structured data-type."
    )
    item_size: int | None = Field(
        default=None, description="Expected size of the struct."
    )
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    endianness: Endianness | None = Field(
        default=None, description="Endianness of structured data type."
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Converts the names, data types, and offsets of the object into a NumPy dtype."""
        dtype_conf = {
            "names": [field.name for field in self.fields],
            "formats": [field.dtype for field in self.fields],
            "offsets": [field.offset for field in self.fields],
        }

        if self.item_size is not None:
            dtype_conf["itemsize"] = self.item_size

        struct_dtype = np.dtype(dtype_conf)  # type: ignore[call-overload]

        if self.endianness is not None:
            struct_dtype = struct_dtype.newbyteorder(self.endianness.symbol)

        return struct_dtype  # type: ignore[no-any-return]
